#ifndef HARKIO_POSITION_HPP_
#define HARKIO_POSITION_HPP_

#include <utility>
#include <cmath>
#include <cerrno>
#include <stdexcept>
#include <boost/operators.hpp>
#include "Vector3.hpp"

/** @file Position.hpp
 * @brief Position class definition.
 */

namespace harkio2
{

/** @brief Represents a position or a direction.
 * Typical unit in this class are meter and radian.
 */
template <typename T>
class Position
    : private boost::equality_comparable<Position<T> >
{
public:
    /** @brief Constructs a null instance.
     */
    Position() throw ();

    /** @brief Constructs an instance representing a direction.
     * This constructor doesn't normalize the given unitVector.
     * @param unitVector A vector representing a direction
     */
    explicit Position(const Vector3<T>& unitVector) throw ();

    /** @brief Constructs an instance representing a position.
     * This constructor doesn't normalize the given unitVector.
     * @param unitVector A vector representing a direction
     * @param length A length between origin and the position
     */
    Position(const Vector3<T>& unitVector, const T& length) throw ();

    /** @brief Returns a unit vector of this instance.
     * @return A unit vector
     */
    const Vector3<T>& GetUnitVector() const throw () { return unitVector; }

    /** @brief Returns a length of this instance.
     * If this instance represents a direction, will return 1.
     * @return A length
     */
    const T& GetLength() const throw () { return length; }

    /** @brief Returns whether this instance represents a position.
     * @return Whether this instance represents a position.
     */
    bool IsPosition() const throw () { return isPosition; }

    /** @brief Returns whether this instance represents a direction.
     * @return Whether this instance represents a direction.
     */
    bool IsDirection() const throw () { return !IsPosition(); }

    /** @brief Returns a position vector.
     * Returns a vector pointing to a position.
     * @return A position vector.
     */
    Vector3<T> GetPositionVector() const throw ();

    /** @brief Returns azimuth angle and elevation angle.
     * Returns a pair of angles.
     * The first element is azimuth angle (theta).
     * The second element is elevation angle (phi).
     * @return a pair of azimuth angle and elevation angle
     */
    std::pair<T, T> GetAngles() const throw (std::range_error);

    /** @brief Checks whether this instance equals to a given instance.
     * @param rhs The other instance
     */
    bool operator ==(const Position& rhs) const throw ();

private:
    /// @brief A unit vector representing the direction.
    Vector3<T> unitVector;

    /** @brief A length between the origin and the position in meters.
     * If isPosition == false, this value represents only
     * the sign of the direction (positive or negative).
     */
    T length;

    /** @brief Whether this instance represents a position.
     * true : this instance represents a position.
     * false : this instance represents a direction.
     */
    bool isPosition;
};

/** @brief Creates an instance representing a direction.
 * This method will normalize 'v'.
 * @param v A vector representing a direction
 * @return An instance representing a direction
 */
template <typename T>
Position<T> MakeDirectionFromVector(const Vector3<T>& v) throw (std::range_error);

/** @brief Creates an instance representing a position.
 * This method doesn't normalize 'v'.
 * @param v A vector representing a position
 * @return An instance representing a position
 */
template <typename T>
Position<T> MakePositionFromVector(const Vector3<T>& v) throw (std::range_error);

/** @brief Creates an instance representing a position.
 * This method will normalize 'v'.
 * @param v A vector representing a direction
 * @param l A length between the origin and the position
 * @return An instance representing a position
 */
template <typename T>
Position<T> MakePositionFromVector(const Vector3<T>& v, const T& l) throw (std::range_error);

/** @brief Creates an instance representing a direction.
 * @param theta A azimuth angle of the direction
 * @param phi A elevation angle of the direction
 * @return An instance representing a directio
 */
template <typename T>
Position<T> MakeDirectionFromPolar(const T& theta, const T& phi) throw ();

/** @brief Creates an instance representing a position.
 * @param theta A azimuth angle of the position
 * @param phi A elevation angle of the position
 * @param r A length between the origin and the position
 * @return An instance representing a position
 */
template <typename T>
Position<T> MakePositionFromPolar(const T& theta, const T& phi, const T& r) throw ();


/* implementation */

namespace
{

/** @brief Normalizes a vector and returns the length of the vector.
 * If the length of the vector is zero,
 * doesn't normalize the vector (so v will not be changed).
 * @param[in,out] v A vector to normalize
 * @param[out] l The length of the vector
 */
template <typename T>
void NormalizeVector(Vector3<T>& v, T& l)
{
    T x = v.GetX();
    T y = v.GetY();
    T z = v.GetZ();
    l = sqrt(x * x + y * y + z * z);
    if (l == T())
    {
        return;
    }
    v = Vector3<T>(x / l, y / l, z / l);
}

template <typename T>
Vector3<T> UnitVectorFromPolar(T theta, T phi)
{
    T cos_phi = cos(phi);
    T theta_ = theta;

    // phi must be in the range [0, pi/2].
    if (cos_phi < 0)
    {
        cos_phi = -cos_phi;
        theta_ = theta + M_PI;
    }

    T x = cos_phi * cos(theta_);
    T y = cos_phi * sin(theta_);
    T z = sin(phi);
    return Vector3<T>(x, y, z);
}

}

template <typename T>
Position<T>::Position() throw ():
    unitVector(), length(), isPosition()
{}

template <typename T>
Position<T>::Position(const Vector3<T>& unitVector) throw ():
    unitVector(unitVector), length(1), isPosition(false)
{}

template <typename T>
Position<T>::Position(const Vector3<T>& unitVector, const T& length) throw ():
    unitVector(unitVector), length(length), isPosition(true)
{}

template <typename T>
Position<T> MakeDirectionFromVector(const Vector3<T>& v) throw (std::range_error)
{
    Vector3<T> v_ = v;
    T length;
    NormalizeVector(v_, length);
    if (length == T())
    {
        throw std::range_error("length of given vector is zero");
    }
    return Position<T>(v_);
}

template <typename T>
Position<T> MakePositionFromVector(const Vector3<T>& v) throw (std::range_error)
{
    Vector3<T> v_ = v;
    T length;
    NormalizeVector(v_, length);
    if (length == T())
    {
        throw std::range_error("length of given vector is zero");
    }
    return Position<T>(v_, length);
}

template <typename T>
Position<T> MakePositionFromVector(const Vector3<T>& v, const T& l) throw (std::range_error)
{
    Vector3<T> v_ = v;
    T length;
    NormalizeVector(v_, length);
    if (length == T())
    {
        throw std::range_error("length of given vector is zero");
    }
    return Position<T>(v_, l);
}

template <typename T>
Position<T> MakeDirectionFromPolar(const T& theta, const T& phi) throw ()
{
    return Position<T>(UnitVectorFromPolar(theta, phi));
}

template <typename T>
Position<T> MakePositionFromPolar(const T& theta, const T& phi, const T& r) throw ()
{
    return Position<T>(UnitVectorFromPolar(theta, phi), r);
}

template <typename T>
Vector3<T> Position<T>::GetPositionVector() const throw ()
{
    T x = unitVector.GetX() * length;
    T y = unitVector.GetY() * length;
    T z = unitVector.GetZ() * length;
    return Vector3<T>(x, y, z);
}

template <typename T>
std::pair<T, T> Position<T>::GetAngles() const throw (std::range_error)
{
    errno = 0;
    T phi = asin(unitVector.GetZ());
    if (errno == EDOM)
    {
        throw std::range_error("domain error: Z of unit vector");
    }

    T cos_phi = cos(phi);
    if (cos_phi == 0 || (unitVector.GetX() == T() && unitVector.GetY() == T()))
    {
        return std::make_pair(T(), phi);
    }

    errno = 0;
    T theta = acos(unitVector.GetX() / cos_phi);
    if (errno == EDOM)
    {
        throw std::range_error("domain error: X of unit vector / cos(phi)");
    }

    if (unitVector.GetY() < 0)
    {
        theta = 2 * M_PI - theta;
    }

    if (cos_phi < 0)
    {
        cos_phi = -cos_phi;
        theta += M_PI;
    }

    return std::make_pair(theta, phi);
}

template <typename T>
bool Position<T>::operator ==(const Position& rhs) const throw ()
{
    return unitVector == rhs.unitVector && length == rhs.length && isPosition == rhs.isPosition;
}

}

#endif // HARKIO_POSITION_HPP_

